import mpi.*;
import pl.edu.agh.icsr.lip.*;
import pl.edu.agh.icsr.janet.tools.*;
import java.text.*;


/*
 Data-ooc.java

 Parallel version of simple irregular loop (see `seq.c' file for details).
 Index arrays are in-core (they fit into the memory so they are not stored on
 disk).
 Data arrays are out of core so they are to be read from disk using MPI IO
 routines
 For simplicity of the code all arrays are assumed to be evenly divisible among
 all nodes.
 */


public class Data_ooc {
    /* size of data array can be evenly divided by number 1...10 (which are allowed
     * total numbers of nodes) */
    final static int N = (2 * 3 * 2); /*5 * 7 * 2 * 9);*/
    final static int L = (2 * N);

    static int rank, size; /* LIP's environment parameters */

    /* ceiling division from HPF 2.0 */
    private static int CD(int j, int k) { return ((j-1)/k+1); }

    public static void main(String[] args) throws MPIException {
	double[] x, y, x_l;
	int[] perm, perm_l, tperm_l;
	int n = N, l = L;
	int n_l,     l_l;    /* numbers of elements on single node (local elements) */
	int n_start, l_start; /* first index on this node */
	
	int i, s, p, ret;
	double sum_y_l, sum_y, sum_x_l, sum_x, z;
	
	mpi.Info info; /* needed by MPI IO */
	mpi.File file;
	long offset;
	mpi.Status status = new mpi.Status();
	
	Datamap datamap, filetype, memtype;
	Schedule schedule;
	IOBufmap bufm;	
	
	MPI.Init(args);
	LIP.Setup(null, 0);
	LIP.DEBUG_create_file();
	info = new mpi.Info();
    
	rank = LIP.Rank();
	size = LIP.Size();
    
	l_l = CD( l, size ); /* number of `x' elements on single node */
	n_l = CD( n, size ); /* number of `y' & `perm' elements on single node */
    
	l_start = l_l * rank;
	n_start = n_l * rank;
    
	/*** Allocate  memory ***/
	
	x = new double[(l_l + n_l)]; /* additional space for non-local
					data (which may be of size `n_l') */
	y = new double[n_l];
	
	x_l    = new double[(l_l + n_l)]; /* additional space for non-local
					     data (which may be of size `n_l') */
    
	perm   = new int[n_l];
	perm_l = new int[n_l];
	tperm_l = new int[n_l];
    
	/*** Fill data  array ***/
    
	z = .1313;
	for (i = 0 ; i < l_start ; i++) /* skip values for previous nodes */
	    z = 4.0 * z * (1.0 - z);
	
	for (i = 0 ; i < l_l ; i++) /* generate values for this node */
	    {
//		x[i] = z - 0.5;
		z = 4.0 * z * (1.0 - z);
		x[i] = l_start+i;
	    }
	
	file = new mpi.File();
	file.open(MPI.COMM_WORLD, "data.file", mpi.File.MODE_RDWR | mpi.File.MODE_CREATE,
		  null);
	file.preallocate(L*8);
	file.set_view(0, MPI.DOUBLE, MPI.DOUBLE, "native", null);
	offset = rank * l_l; /* Each node writes its part of a file */
	file.write_at(offset, new VDouble(x),l_l,MPI.DOUBLE,status);

	dumpDoubleArray(x, "Array x", 0, l_l);
  
	/*** Fill index array ***/

	/* skip values for previous nodes */
	for (p = (int)(.1313 * l), s = -l, i = 0 ; i < n_start ; i++, s += 2)
	    p = (p + s + l) % l;
	
	for (i = 0 ; i < n_l ; i++, s += 2) /* generate values for this node */
	    {
		perm[i] = p;
		p = (p + s + l) % l;
	    }

	dumpIntArray(perm, "Index array perm", 0, n_l);
    
	/*** Inspector  phase ***/

	/* create data map which describes how data array `x' is partitioned */
	datamap = new Datamap(new Datamap.BlockDistribution(l, l_l), 1, 0);
    
	/* create IO bufmap which maps local data buffer on the data array on disk */  
	bufm = new IOBufmap(l_l);
    
	schedule = new Schedule(l_l);

	LIP.Localize(datamap, 
		     new LIP.ContIntIndexer(perm), 
		     new LIP.ContIntIndexer(perm_l),
		     0, schedule);

	dumpIntArray(perm_l, "Local index array perm_l", 0, n_l);

	LIP.OOC_Localize(new LIP.ContIntIndexer(perm_l),
			 new LIP.ContIntIndexer(tperm_l),
			 0, new VDouble(x_l), MPI.DOUBLE, bufm, schedule);

	dumpIntArray(tperm_l, "Transformed local index array tperm_l", 0, n_l);

	datamap.free(0);
	datamap = null;

	/* schedule has to be committed before its use in communication functions */
	schedule.commit();
  
	/* get datatypes mapping memory to disk */
	IndicesDatatypes id;
	try {
	    id = bufm.get_datatype(MPI.DOUBLE, LIP.DATA_NEW);
	    file.set_view(l_l*rank*8, MPI.DOUBLE, id.filetype, "native", info);
	    file.read_at(0, new VDouble(x_l), 1, id.memtype, status);
	} catch (LipException e) {
	    System.out.println("rank " + rank + ": gotcha!!!"); System.out.flush();
	    file.set_view(l_l*rank*8, MPI.DOUBLE, MPI.DOUBLE, "native", info);
	}

	dumpDoubleArray(x_l, "Local array x_l read from disk", 0, l_l);

	/*** Executor   phase ***/

	/* gather non-local data from other nodes */
	LIP.Gather(new VDouble(x_l), new VDouble(x_l).subArray(l_l), MPI.DOUBLE, schedule );

	dumpDoubleArray(x_l, "Local array x_l with Ghost area", 0, l_l + n_l);

        /* irregular loops */

	for (i = 0 ; i < n_l ; i++)
	    y[i] = -x_l[ tperm_l[i] ];
	
	/* clear the ghost area */
	for (i = 0 ; i < n_l ; i++)
	    x_l[ i + l_l ] = 0.0;
	
	for (i = 0 ; i < n_l ; i++)
	    x_l[tperm_l[i] ] += y[i];
	
	
	
	/* scatter non-local data to other nodes */
	LIP.Scatter(new VDouble(x_l).subArray(l_l), new VDouble(x_l), MPI.DOUBLE, schedule, 
		    LIP.OP_SUM);
	
	schedule.free();
	schedule = null;

	/* modify data values on the disk*/
	id = bufm.get_datatype(MPI.DOUBLE, LIP.DATA_ALL);
	file.set_view(l_l*rank*8, MPI.DOUBLE, id.filetype, "native", info);
	file.write_at(0, new VDouble(x_l), 1, id.memtype, status);
	
	/* read modified x array */  
	file.set_view(0, MPI.DOUBLE, MPI.DOUBLE, "native", null);
	offset = rank * l_l; /* Each node writes its part of a file */
	file.read_at(offset, new VDouble(x), l_l, MPI.DOUBLE, status);
	
    
    
	if (0 == rank && size > 1) {
	    System.out.println("" + size + " computing nodes. " +
			       "Each has " + n_l + " `y' elements and " +
			       l_l + " `x' elements.");
	}
    
	for (sum_y_l = 0.0, i = 0 ; i < n_l ; i++)
	    sum_y_l += y[i];

	
	double[] asum_y = new double[1];
	MPI.COMM_WORLD.Reduce(new double[] { sum_y_l }, 0,
                              asum_y, 0,
			      1, MPI.DOUBLE, MPI.SUM, 0);
    
	sum_y = asum_y[0];
	if (0 == rank)
	    System.out.println("The sum of " + n + " elements of `y' is " + sum_y);

	for (sum_x_l = 0.0, i = 0 ; i < l_l ; i++)
	    sum_x_l += x[i];

	double[] asum_x = new double[1];
	MPI.COMM_WORLD.Reduce(new double[] { sum_x_l }, 0, 
                              asum_x, 0,
			      1, MPI.DOUBLE, MPI.SUM, 0);
	
	sum_x = asum_x[0];
	if (0 == rank)
	    System.out.println("The sum of " + l + " elements of `x' is " + sum_x);

	file.close();


    /*    info = null;
	  file = null;
    status = null;
    System.gc();
    System.runFinalization();
   
    System.out.println("test8");*/
	LIP.Exit();
	MPI.Finalize();
    }

    private static void dumpIntArray(int[] arr, String desc, int beg, int end) {
	StringBuffer buf = new StringBuffer();
	buf.append("Rank " + rank + " - " + desc + ":\n");
	for (int i=beg; i<end; i++)
	    buf.append("" + arr[i] + " ");
	buf.append("\n\n");
	System.out.print(buf);
	System.out.flush();
    }

    private static void dumpDoubleArray(double[] arr, String desc, int beg, int end) {
	StringBuffer buf = new StringBuffer();
	buf.append("Rank " + rank + " - " + desc + ":\n");
	for (int i=beg; i<end; i++)
	    buf.append(MessageFormat.format("{0,number,#0.000000} ", 
					    new Double[] { new Double(arr[i]) }));
	buf.append("\n\n");
	System.out.print(buf);
	System.out.flush();
    }

}

