/* 
LIP Library
File iobufmap.c -contains functions for creation and menagemnt
of object that stores information about mapping disc to memory
created by Tomasz Baranski 27.11.1999
*/

#include <stdlib.h>
#include <lipimpl.h>
#include <liptools.h>
#include <hashsort.h>

/*  Function takes two hashes minuend_hash and subtrahend_hash
	and returns entries in the first one that are not in the second one */
	
int 
LIP_hash_difference(
	int * minuend_hash_key,
	int * minuend_hash_val,
	int * subtrahend_hash_key,
	int * subtrahend_hash_val,
	int size,	/* size of hash tables */
	int * memplaces,/* an array of memory indices that should be changed */
	int * discplaces,/* an array of disc indices that should be changed */
	int * count /* number of significant entries in memplaces[] and discplaces[] */
)
{
  int i, index, empty,ret;
  
  *count=0; /* initiate couter */
  
  for (i=0;i<size;i++)
  {
    if (minuend_hash_val[i] != HASH_EMPTY)
    {
      /* find out whether entry minuend_hash_key[i] exists in
         subtrahend_hash (operation hash_query) */
      ret=hash_query(minuend_hash_key[i],&index,&empty,subtrahend_hash_val,
      		subtrahend_hash_key,size);
      
      if(LIP_SUCCESS!=ret){
        return ret;
      }
      
      if(empty)
      {
        memplaces[*count] = minuend_hash_val[i];
        discplaces[*count] = minuend_hash_key[i];
        (*count)++;
      }
    }
  }
  return LIP_SUCCESS;
} /* LIP_hash_difference*/



/*  Function that puts information in hash_key and hash_val
	to the arrays memplaces[] (for hash_val) and discplaces[] (for hash_key),
	so that are no empty entries in output arrays.
	The number of this entries is stored in the count variable */

int 
LIP_hash_significant_entries(
	int * hash_key,
	int * hash_val,
	int size,		  /* size of hash tables */
	int * memplaces,  /* an array of memory indices that should be changed */
	int * discplaces, /* an array of disc indices that should be changed */
	int * count       /* number of significant entries in memplaces[] 
						 and discplaces[] */
)
{ 
  int i;

  *count = 0;
  
  for (i=0;i<size;i++)
  {
    if (hash_val[i] != HASH_EMPTY)
    {
      memplaces[*count] = hash_val[i];
      discplaces[*count] = hash_key[i];
      (*count)++;
    }
  }
  return LIP_SUCCESS;
} /* LIP_hash_significant_entries*/



/*  Function that creates a new MPI data type which may be used in MPI_IO
	routines for reading or writing data.
	The flags argument can combine the following constants:
	LIP_DATA_OLD - then the returned MPI data type will refer to the old
		data elements (which should be written out to disc)
	LIP_DATA_NEW - then the returned MPI data type will refer to the new
		data elements (which should be read in from disc)
	LIP_DATA_ALL -then the returned MPI data type will refer to all data
		elements in the memory buffer - useful when the user wants to
		write all the data after performing the last calculation
  Function returns 0 when there is no data to exchange -
  - no MPI datatypes are created!
*/


int
LIP_IOBufmap_get_datatype(LIP_IOBufmap bmap,
					MPI_Datatype datatype,/*type of stored data*/
                    MPI_Datatype *memtype,
                    		/* MPI_Datatype that describes
                            indices of data in memory (used
                            in MPI_File_write_* and
                            MPI_File_read_* functions) */
					MPI_Datatype *filetype,
                    		/* MPI_Datatype that describes
                            indices of data on disc
                            (used in MPI_File_set_view) */
					int flag)
{
  LIP_IIOBufmap *ibmap;
  int size, count;
  int *memplaces, *discplaces;
  
  int i,ret;
  
  char* pname="LIP_IOBufmap_get_datatype";
  
  ibmap = LIP_IOBufmap_convert(bmap);
  if(NULL==ibmap){
   LIP_Error_ret(pname,"bad IOBufmap",LIP_INVALID_ARGUMENT);
   return LIP_INVALID_ARGUMENT;
   
  }
  size = ibmap->htab_size;
  

/*(1)Allocate memory for array that stores indices in memory of data 
     beeing moved -memplaces[]*/
  memplaces =(int*) malloc(size * sizeof(*memplaces));

  if (NULL==memplaces){
    LIP_Error_ret(pname,NULL,LIP_OUT_OF_MEMORY);
    return LIP_OUT_OF_MEMORY;
  }
 
/*(2)Allocate memory for array that stores indices on disc of data 
     beeing moved -discplaces[]*/
  discplaces = (int *) malloc(size * sizeof(*discplaces));
  if (NULL==discplaces){
    LIP_Error_ret(pname,NULL,LIP_OUT_OF_MEMORY);
    return LIP_OUT_OF_MEMORY;
  }
     
/*(3.1)if we want to write all data */
  if (flag & LIP_DATA_ALL)
    ret=LIP_hash_significant_entries(ibmap->newhash_key,ibmap->newhash_val,
    							size,memplaces,discplaces,&count);
                               
/*(3.2)if we want only old data (to read) */
  else
  if (flag & LIP_DATA_OLD)
  	ret=LIP_hash_difference(ibmap->oldhash_key,ibmap->oldhash_val,
  					ibmap->newhash_key,ibmap->newhash_val,
					size,memplaces,discplaces,&count);
                      
/*(3.3)if we want only new data (to read) */
  else
  if (flag & LIP_DATA_NEW)
  	ret=LIP_hash_difference(ibmap->newhash_key,ibmap->newhash_val,
  					ibmap->oldhash_key,ibmap->oldhash_val,
					size,memplaces,discplaces,&count);
  
  
 if(ret!=LIP_SUCCESS){
  LIP_Error_ret(pname,"error ",ret);
  return ret;
 
 } 
 if (count ==0)
 {
       
    LIP_Type_create_indexed_block(count,1,discplaces,datatype,filetype);
    MPI_Type_commit(filetype);

    LIP_Type_create_indexed_block(count,1,memplaces,datatype,memtype);
    MPI_Type_commit(memtype);

      free(discplaces);
      free(memplaces);
  /*    *filetype = MPI_DATATYPE_NULL;
      *memtype = MPI_DATATYPE_NULL;  */
      return LIP_SUCCESS;
 }
/*(4)Sort discplaces[] and memplaces[] to have
     the entries in discplaces[] in increasing order
     using quicksort */

  hashsort(count, discplaces, memplaces);

/*(5)Create data using MPI*/ 

  LIP_Type_create_indexed_block(count,1,discplaces,datatype,filetype);
  MPI_Type_commit(filetype);

  LIP_Type_create_indexed_block(count,1,memplaces,datatype,memtype);
  MPI_Type_commit(memtype);


/*(6)Free memory */
 free(memplaces);
 free(discplaces);
 return LIP_SUCCESS;

} /* LIP_IOBuffer_get_datatype*/

/* function that returns the pointer to the IOBufmap structure
   that has the handle bmap
   input : handle
   output: pointer to the structure */
   

static LIP_IIOBufmap * iobufmap_array[MAX_IOBUFMAP];
static int iobufmap_index=0;
/*inits handler array*/
int LIP_IOBufmap_array_init(){
 int i;
 for(i=0;i<MAX_IOBUFMAP;i++){
  iobufmap_array[i]=NULL;
 }
 return LIP_SUCCESS;
}

/*frees handler array*/
int LIP_IOBufmap_array_free(){
 int i;
 LIP_IOBufmap bm;

 for(i=0;i<MAX_IOBUFMAP;i++){
  if(iobufmap_array[i]!=NULL){
   bm=i+1;
   LIP_IOBufmap_free(&bm);
  }
 }
 return LIP_SUCCESS;
}
  
LIP_IIOBufmap *
LIP_IOBufmap_convert(LIP_IOBufmap bmap)				/* KZ & TB */
{
 /*if bmap is a handle of the NULL structure return NULL
   else return pointer from the array of pointers to all created 
   structures */

  return LIP_IOBUFMAP_NULL == bmap ? NULL : iobufmap_array[bmap - 1];
}



/* function for creation of a new IOBufmap
   input:  size of memory buffer,
   output: handle of the structure (*bufm)
   information if the operation was successful or not (value returned)
*/
int
LIP_IOBufmap_create(int size, LIP_IOBufmap *bufm)      /* KZ & TB */
{
 /* pointer to the structure beeing created */
  LIP_IIOBufmap *bm;

  int i,found;
  char*pname="LIP_IOBufmap_create";
 /* Check if size <= 0 */
   if (size <= 0){
     LIP_Error_ret(pname,"size < 0",LIP_INVALID_ARGUMENT); 
     return LIP_INVALID_ARGUMENT;
   }

 /* Search the array iobufmap_array[] for an empty place  
    if there is no place return LIP_ERROR;
    else iobufmap_index is the index that marks the place for new structure */ 
  
  if (iobufmap_array[iobufmap_index] != NULL)
   {      
    found=0;
    for(i=0;i<MAX_IOBUFMAP;i++){
         if(iobufmap_array[i] == NULL){
            iobufmap_index = i;
	    found=1;
            break;
	 } 
     }
     if (!found ){
       LIP_Error_ret(pname,"MAXIOBUFMAP is too small",LIP_TOO_MANY_HANDLES);
       return LIP_TOO_MANY_HANDLES;
     
     } 
  }
 /* Allocation of the memory for the new structure */ 
  
  iobufmap_array[iobufmap_index] = bm = malloc( sizeof *bm );
 
 /* In case of unsuccessful allocation */
  if (NULL == bm){
     LIP_Error_ret(pname,NULL,LIP_OUT_OF_MEMORY);
     return LIP_OUT_OF_MEMORY;
  }
 /* Allocation is successful */
 /* Creation of the structure */
  bm->size  = size;
  bm->oldhash_key = NULL;
  bm->oldhash_val = NULL;
  bm->newhash_key = NULL;
  bm->newhash_val = NULL;
  bm->htab_size=0;

 /* handles start from 1; 0 is NULL structure*/
  *bufm = ++iobufmap_index;
  iobufmap_index=iobufmap_index %MAX_IOBUFMAP;

  return LIP_SUCCESS;

} /* LIP_IOBufmap_create */


/* function for deleting structure 
   input:  handle of the structure (*bufm)
   output: information if the operation was successful (value returned)*/

int
LIP_IOBufmap_free(LIP_IOBufmap *bufm)
{
  int i;
  char* pname="LIP_IOBufmap_free";
  /* Check if IOBufmap isn't NULL*/
  if (LIP_IOBUFMAP_NULL == *bufm){
    LIP_Error_ret(pname,"trying to free NULL IOBufmap ",LIP_INVALID_ARGUMENT);
    return LIP_INVALID_ARGUMENT;
  }
 /* find the structure corresponding to input handle */
   
  i = *bufm -1;

  if (iobufmap_array[i] == NULL){
    LIP_Error_ret(pname,"trying to free NULL IOBufmap",LIP_INVALID_ARGUMENT);
    return LIP_INVALID_ARGUMENT;
  } 
 /* free resources for that structure
       size of all hashes is bm->htab_size
       size of index_to_change is bm->changes */

  if (iobufmap_array[i]->oldhash_key){
    free(iobufmap_array[i]->oldhash_key);
    iobufmap_array[i]->oldhash_key=NULL;
  }
  if (iobufmap_array[i]->oldhash_val){
    free(iobufmap_array[i]->oldhash_val);
    iobufmap_array[i]->oldhash_val=NULL;
  }
  if (iobufmap_array[i]->newhash_key){
    free(iobufmap_array[i]->newhash_key);
    iobufmap_array[i]->newhash_key=NULL;
  }    
  if (iobufmap_array[i]->newhash_val){
    free(iobufmap_array[i]->newhash_val);
    iobufmap_array[i]->newhash_val=NULL;
  }
  free(iobufmap_array[i]);
  iobufmap_array[i] = NULL;
  
 /*(4) change handle to 0 (means this IOBufmap is NULL)*/
  *bufm = LIP_IOBUFMAP_NULL;

  return LIP_SUCCESS;
} /* LIP_IOBufmap_free */




