/*
 LIP library 
 localize.c -contains function for translation non-local indices
 to indices pointing to the ghost area of processor's memory buffer 
 it also contains functions for communication schedule menagement
 created by Piotr Luszczek
 */

#include <stdlib.h>
#include <string.h>

#include <lipimpl.h>
#include <liptools.h>
#include <debug.h>
#define p(x) printf("[%02d]" __FILE__ "(%d)" #x "=%g\n",LIP_env.rank,__LINE__,(double)(x)),fflush(0)
#define pa(a,i) printf("[%02d]" __FILE__ "(%d)" #a "[%d]=%g\n",LIP_env.rank,__LINE__,(i),(double)((a)[i])),fflush(0)

static int suma=0;
static int cs=0;
int
LIP_Localize(LIP_Datamap datamap, void *iglobal, LIP_Indextype igtype,
  void *ilocal, LIP_Indextype iltype, int count, int flags,
  LIP_Schedule *schedule, LIP_Indexinfo *info)
{
  int i, j, sumka=0,sumkap=0;
  int *g_indcs, *l_indcs;
  int data_size, data_size_1;
  int *procs;
  int *location;
  int local;
  int *shift, rdispl;

  int h_ind;
  int *htab_vals, *htab_keys, htab_size;
  int found,*ret;
  int nothing;
  LIP_ISchedule *s;
  char*pname="LIP_Localize";

  /* FIXME: only contiguous integer C indexing supported for now */
  if (igtype != LIP_INDEXTYPE_INT || iltype != LIP_INDEXTYPE_INT){
      LIP_Error_quit(pname,"indextype ",LIP_NOT_SUPPORTED);
  }
  g_indcs = iglobal;
  l_indcs = ilocal;

  /* if there is only one node */
  if (1 == LIP_env.size)
  {
    /* The algorithm below will probably survive when there is only one node.
    ** However it will perform a lot of unnecessary computation (mostly long
    ** loops over arrays). */

    memcpy( l_indcs, g_indcs, count * (sizeof *l_indcs) );

    s = LIP_Schedule_convert( *schedule );

    if(NULL==s){
     LIP_Error_ret(pname,"invalid schedule",LIP_INVALID_ARGUMENT);
     return LIP_INVALID_ARGUMENT;
    }
    if(NULL==s->esizes){
    s->esizes= malloc(sizeof(*s->esizes));
    if(NULL==s->esizes){
     LIP_Error_ret(pname,NULL,LIP_OUT_OF_MEMORY);
     return LIP_OUT_OF_MEMORY;
    }
    }
    if (NULL==s->edispls){
    s->edispls= malloc(sizeof(*s->edispls));
    if(NULL==s->edispls){
     LIP_Error_ret(pname,NULL,LIP_OUT_OF_MEMORY);
     return LIP_OUT_OF_MEMORY;
    }
    }
    if(NULL==s->isizes){
    s->isizes= malloc(sizeof(*s->isizes));
    if(NULL==s->isizes){
     LIP_Error_ret(pname,NULL,LIP_OUT_OF_MEMORY);
     return LIP_OUT_OF_MEMORY;
    }
    }
    if(NULL==s->idispls){
    s->idispls= malloc(sizeof(*s->idispls));
    if(NULL==s->idispls){
     LIP_Error_ret(pname,NULL,LIP_OUT_OF_MEMORY);
     return LIP_OUT_OF_MEMORY;
    }
    }
    s->edispls[0]=0;
    s->isizes[0]=0;
    s->esizes[0]= 0;
    s->idispls[0]= 0;
    
    return LIP_SUCCESS;
  }

  /*
  Algorithm outline:
  - hash table is created (for finding duplicated indices)
  - all the global indices are "dereferenced" i.e. home processor and location
    on the home processor is found
  - for each global index:
    if it is local then its local index is inserted into 'l_indcs'
    if it is non-local then
      - if it is not in hash table then it is inserted to hash table and new
      position in 'ghost area' is associated with the index; placement in
      'ghost area' is very regular - the area is split into blocks - each block
      store indices from particular processor
      - if it is already in hash table then its local placement (in
      'ghost area') is taken form hash table and inserted into 'l_indcs'
   */

  s = LIP_Schedule_convert( *schedule );
  data_size = s->data_size;
  data_size_1 = data_size - 1;

  procs = malloc( (sizeof *procs) * count );

  LIP_Assert(procs,pname,"possibly too big array of indices ",LIP_OUT_OF_MEMORY);
  if(NULL==s->isizes){
   s->isizes = malloc( (sizeof *s->isizes) * LIP_env.size );
   LIP_Assert(s->isizes,pname,"while filling schedule",LIP_OUT_OF_MEMORY);
  }
  if(NULL==s->idispls){
   s->idispls = malloc( (sizeof *s->idispls) * LIP_env.size );

   LIP_Assert(s->idispls,pname,"while filling schedule",LIP_OUT_OF_MEMORY);
  }
  
  if(LIP_SUCCESS!=LIP_Dereference( datamap, iglobal, igtype, ilocal, iltype, procs, count,
    LIP_DEREF_ALL, s->isizes )){
    LIP_Error_quit(pname,"error in LIP_Dereference",LIP_ERROR);
  }
  
  local = s->isizes[LIP_env.rank];
  s->isizes[LIP_env.rank] = 0;
  
  for (s->idispls[0] = 0, i = 1 ; i < LIP_env.size ; i++)
  {
    s->idispls[i] = s->idispls[i - 1] + s->isizes[i - 1];
    s->isizes[i - 1] = 0; /* these values will be reconstructed while creating
    			  ** 'location' array, but with duplicates removed, so
			  ** values reconstructed will probably be less than
			  ** the ones returned from LIP_Dereference() */
  }
  s->isizes[LIP_env.size - 1] = 0; /* the loop above skips the last element */

  /* permutation(valid non-localy, i.e. it can be used to reference data array
  ** on other nodes) array of all non-local indices; its size is exaggerated
  ** since duplicate indices hasn't been removed yet */

  location = malloc( (sizeof *location) * (count - local) );
  LIP_Assert(location,pname,"possibly too many non-local indices",LIP_OUT_OF_MEMORY);


  /* FIXME: what if there are no non-local indices - no hash table needed */
  /* hash table will store only non-local references and will be 50% larger in
   * size than number of these references (plus prime-round-up) */
  htab_size = prime_round_up( (count - local) * 1.5 );
  
  /* check return value */
  htab_keys = malloc( (sizeof *htab_keys) * htab_size );
  LIP_Assert(htab_keys,pname,"possibly too many non-local indices",LIP_OUT_OF_MEMORY);
  
  htab_vals = malloc( (sizeof *htab_vals) * htab_size );
  LIP_Assert(htab_vals,pname,"possibly too many non-local indices",LIP_OUT_OF_MEMORY);
  /* FIXME: faster filling could be implemented */
  for (i = htab_size - 1 ; i >= 0 ; i--)
  {
    htab_vals[i] = data_size_1; /* `data_size_1' is a special value which means
                                       that position is empty */
    htab_keys[i] = -1;
  }

  /* FIXME: reverse order could be faster */
  /* go through all indices and "localize" them */
  for (i = 0 ; i < count ; i++)
  {
    /* if this is non-local index */
    if (procs[i] != LIP_env.rank)
    {
      sumkap++;
      /* compute hash index; for best results `htab_size' should be prime
       * number */
      h_ind = g_indcs[i] % htab_size;

      found = 0; /* initially the global index is not found */

      /* if postion that was found in hash table is empty */
      if (htab_vals[h_ind] == data_size_1)
        ; /* nothing for now - things are done later; `h_ind' stores index */
      /* if position that was found in hash table contains index being sought*/
      else if (htab_keys[h_ind] == g_indcs[i])
        found = 1;
      /* entry that was found in hash table is _not_ empty */
      else if (data_size_1 != htab_vals[h_ind])
      {
        j = -htab_size;
        /* look for empty position */
        do
        {
          j += 2;

          h_ind += ABS(j);
          if (h_ind >= htab_size)
            h_ind -= htab_size;
        } while (htab_keys[h_ind] != g_indcs[i] &&
                 htab_vals[h_ind] != data_size_1 && j < htab_size);

	/* what if impossible happens: hash table is full? */
	
        if (j >= htab_size){
	  LIP_Error_quit(pname,"hash table is full",LIP_ERROR);
	}

	if (htab_keys[h_ind] == g_indcs[i])
	  found = 1;
      }

      /* if entry for `g_indcs[i]' was not found in hash table, then add it to
       * the import locations; set values in hash table at position `h_ind' */
      if (! found)
      {
        sumka++;
	htab_keys[h_ind] = g_indcs[i];

	/* procs[i] = node on which data with index 'i' resides
	** idispls[i] = displacement in 'location' (where the buffer starts for
	**   data (permutation indices)) for processor 'i'
	** isizes[i] = current position in buffer of data which will be sent to
	**   node 'i' (it is succesfully incremented to yield the number of
	**   elements that should be sent to node 'i') */

	/* new position in a block of indices stored on node 'procs[i]' */
	j = s->idispls[ procs[i] ] + s->isizes[ procs[i] ];
        s->isizes[ procs[i] ]++; /* position for next index */

	/* location on foreign node */
        location[j] = l_indcs[i];

	/* now it points to buffer's ghost area */
	htab_vals[h_ind] = l_indcs[i] = data_size + j;
    
      }
      else{
	  /*copy postion that was already assigned*/
	  l_indcs[i] = htab_vals[h_ind];
      }
    }
  }
/*#ifdef DEBUG 
    for (i = 0 ; i < count ; i++)
      fprintf(debf, "%d:%d, %d\n", i, g_indcs[i], l_indcs[i]);
#endif
*/
  if(NULL==s->esizes){
   s->esizes  = malloc( (sizeof *s->esizes) * LIP_env.size );
   LIP_Assert(s->esizes,pname,"while filling schedule",LIP_OUT_OF_MEMORY);
  }
  
  if(NULL==s->edispls){
   s->edispls = malloc( (sizeof *s->edispls) * LIP_env.size );
   LIP_Assert(s->edispls,pname,"while filling schedule",LIP_OUT_OF_MEMORY);
  }
  /* `esizes' is copied from other nodes' `isizes' array, therfore information
   * stored in the former array is reduntant but with this redundancy
   * no communication is required whenever `esizes' is to be used */
  MPI_Alltoall( s->isizes, 1, MPI_INT, s->esizes, 1, MPI_INT, LIP_env.comm );

  /* create array of export displacements */
  for (s->edispls[0] = 0, i = 1 ; i < LIP_env.size ; i++)
    s->edispls[i] = s->edispls[i - 1] + s->esizes[i - 1];
  
  s->elocations = realloc( s->elocations,(sizeof *s->elocations) *
    (s->esizes[LIP_env.size - 1] + s->edispls[LIP_env.size - 1]) );

  if(0< (s->esizes[LIP_env.size - 1] + s->edispls[LIP_env.size - 1])){
  
  LIP_Assert(s->elocations,pname,"too many indices to export",LIP_OUT_OF_MEMORY);
 }  
  
  MPI_Alltoallv( location, s->isizes, s->idispls, MPI_INT, s->elocations,
    s->esizes, s->edispls, MPI_INT, LIP_env.comm );

  shift = malloc( (sizeof *shift) * LIP_env.size );
  LIP_Assert(shift,pname,NULL,LIP_OUT_OF_MEMORY);
 
  /* create array of import displacements; they have to be created once again
   * because in the previous creation duplicate values has not been taken into
   * account, so probably (further optimization could check whever duplicate
   * references to data array occured, however this optimization would not gain
   * much) information stored in `idspls' has changed; the change of import
   * displacements affects also `localized' indices (i.e. indices which pointed
   * to non-local data elements and has been changed to point to ghost area);
   * they have to shifted (substracted constant value) which depends on shift
   * of blocks in ghost area which in turn was caused by eliminating duplicate
   * indices
   */
  /* s->idispls[0] was already set to zero */
  for (rdispl = shift[0] = 0, i = 1 ; i < LIP_env.size ; i++)
  {
    rdispl += s->isizes[i - 1];
    shift[i] = s->idispls[i] - rdispl;

    s->idispls[i] = rdispl;
  }

  /* for each 'localized' index */
  for (i = count - 1 ; i >= 0 ; i--)
    /* if it points to non-local data */
    if (procs[i] != LIP_env.rank)
      /* shift the index */
      l_indcs[i] -= shift[ procs[i] ];

  /*
  for (s->idispls[0] = 0, i = 1 ; i < LIP_env.size ; i++)
    s->idispls[i] = s->idispls[i - 1] + s->isizes[i - 1];
  */

  /*
  *gs_area = s->isizes[LIP_env.size - 1] + s->idispls[LIP_env.size - 1];
  */

  free( shift );

  free( location );

  free( htab_keys );
  free( htab_vals );

  free( procs );
  suma+=sumka;
  cs+=count;
  return LIP_SUCCESS;
} /* LIP_Localize */

static LIP_ISchedule *schedule_array[MAX_SCHEDULE];
static int schedule_index = 0;
int LIP_Schedule_array_init(){
 int i;
 for(i=0;i<MAX_SCHEDULE;i++){
  schedule_array[i]=NULL;
 } 
 return LIP_SUCCESS;
}
int LIP_Schedule_array_free(){
 int i;
 LIP_Schedule sch;
 for(i=0;i<MAX_SCHEDULE;i++){
  if(schedule_array[i]!=NULL){
   sch=i+1;
   LIP_Schedule_free(&sch);
  }
 } 
 return LIP_SUCCESS;
}
/*
 size - the size of local data array
 */
int
LIP_Schedule_create(int size, LIP_Schedule *schedule)
{
  int i,found;
  LIP_ISchedule *s;

  /* local array has to have some size; also LIP_Localize() uses value:
   * `size' - 1 (taken from schedule) if:
   * 1. LIP_Localize() is implemented with unsigned arithmetic and
   * 2. `size' <= 0 then
   * LIP_Localize() may use strange values for schedule generation. It probably
   * will be OK, but... */
  char*pname="LIP_Schedule_create";
  
  if (size <= 0){
     LIP_Error_ret(pname,"size < 0 !",LIP_INVALID_ARGUMENT);
     return LIP_INVALID_ARGUMENT;
  }
  
  if(NULL!=schedule_array[schedule_index]){
   found=0;
   for(i=0;i<MAX_SCHEDULE;i++){
    if (NULL==schedule_array[i]){
     schedule_index=i;
     found=1;
     break;
    }
   }
   if(!found){
     LIP_Error_ret(pname,"MAX_SCHEDULE is too small",LIP_TOO_MANY_HANDLES);
     return LIP_TOO_MANY_HANDLES;
   }
  }
  
  /*schedule creation*/
  schedule_array[schedule_index] = s = malloc( sizeof *s );
  
  if (NULL == s){
     *schedule=LIP_SCHEDULE_NULL;
     LIP_Error_ret(pname,"for creating schedule",LIP_OUT_OF_MEMORY);
     return LIP_OUT_OF_MEMORY;
  }
  
  s->esizes  = NULL;
  s->edispls = NULL;
  s->isizes  = NULL;
  s->idispls = NULL;
  s->elocations = NULL;
  s->data_size = size;
  s->committed = 0;
  s->user_buf = NULL;

  /* `schedules' start from 1; 0 is NULL schedule */
  *schedule = ++schedule_index;
  schedule_index=schedule_index % MAX_SCHEDULE;
  return LIP_SUCCESS;
} /* LIP_Schedule_create */

int
LIP_Schedule_free(LIP_Schedule *schedule)
{
  LIP_ISchedule* s;
  char*pname="LIP_Schedule_free";
  
  if (LIP_SCHEDULE_NULL == *schedule){
    LIP_Error_ret(pname,"free NULL schedule",LIP_INVALID_ARGUMENT);
    return LIP_INVALID_ARGUMENT;
  }
  s=LIP_Schedule_convert(*schedule);
  if(NULL==s){
    LIP_Error_ret(pname,"bad schedule handle",LIP_INVALID_ARGUMENT);
    return LIP_INVALID_ARGUMENT;
  }
  /*free resources */
  if (NULL!=s->esizes){
      free(s->esizes);
      s->esizes=NULL;
  }
  if (NULL!=s->edispls){
      free(s->edispls);
      s->edispls=NULL;
  }
  if (NULL!=s->isizes){
      free(s->isizes);
      s->isizes=NULL;
  }
  if (NULL!=s->idispls){
      free(s->idispls);
      s->idispls=NULL;
  }  
  if (NULL!=s->elocations){
      free(s->elocations);
      s->elocations=NULL;
  }
  if (NULL!=s->user_buf){
      free(s->user_buf);
      s->user_buf=NULL;
  }
  free(s);
  
  schedule_array[*schedule-1]=NULL;
  *schedule = LIP_SCHEDULE_NULL;  
  return LIP_SUCCESS;
} /* LIP_Schedule_free */

LIP_ISchedule *
LIP_Schedule_convert(LIP_Schedule s)
{
  return LIP_SCHEDULE_NULL == s ? NULL : schedule_array[s - 1];
}

int
LIP_Schedule_commit(LIP_Schedule *schedule)
{
  LIP_ISchedule *is;
  char*pname="LIP_Schedule_commit";
  is = LIP_Schedule_convert( *schedule );

  if (NULL == is){
    LIP_Error_ret(pname,"bad schedule handle",LIP_INVALID_ARGUMENT);
    return LIP_INVALID_ARGUMENT;
  }
  /* FIXME: more elaborate schedule committing is needed for more elaborate
   * parallelizing schemes */
  is->committed = 1;

  return LIP_SUCCESS;
} /* LIP_Schedule_commit */
